structure Eval = struct
  (* Standard evaluation *)			 
  open NRC
  open Value

  exception EvalFailure

  fun eval ctx exp = 
      let fun ev (Var(x)) = (case Var.Map.find(ctx,x) 
			      of SOME v => v)
	    | ev (Let(x,e1,e2)) = 
	      let val v = ev e1 
	      in eval (Var.Map.insert(ctx,x,v)) e2
	      end

	    | ev (BoolC c) = BoolV c
	    | ev (And(e1,e2)) = 
	      (case (ev e1, ev e2) 
		of (BoolV b1, BoolV b2) => BoolV (b1 andalso b2))
	    | ev (Neg(e)) = 
	      (case (ev e) 
		of (BoolV b) => BoolV (not(b)))
	    | ev (IfThenElse(e,e1,SOME e2)) = 
	      (case (ev e) 
		of (BoolV b) => if b then ev e1 else ev e2)
	    | ev (IfThenElse(e,e1,NONE)) = 
	      (case (ev e) 
		of (BoolV b) => if b then ev e1 else SetV([]))
	    | ev(Equal(e1,e2)) = BoolV(Value.eq(ev e1, ev e2))
	       	       
	    | ev (IntC c) = IntV c
	    | ev (Plus(e1,e2)) = 
	       (case (ev e1, ev e2) 
		of (IntV i1, IntV i2) => IntV (i1 + i2))
	    | ev (Times(e1,e2)) = 
	       (case (ev e1, ev e2) 
		of (IntV i1, IntV i2) => IntV (i1 * i2))
	    | ev (Min(e)) = 
	      let val SOME maxInt = Int.maxInt 
	      in(case (ev e) 
		  of (SetV vs) => IntV (List.foldr (fn (IntV i, s) => Int.min(i,s)) maxInt vs))
	      end
	    | ev (Max(e)) = 
	      let val SOME minInt = Int.minInt 
	      in 
	       (case (ev e) 
		of (SetV vs) => IntV (List.foldr (fn (IntV i, s) => Int.max(i,s)) minInt vs))
	       end
	    | ev (Sum(e)) = 
	       (case (ev e) 
		of (SetV vs) => IntV (List.foldr (fn (IntV i, s) => i+s) 0 vs))
	      
	    | ev (Empty(e)) = 
	      (case (ev e) 
		  of (SetV []) => BoolV true
		   | (SetV _) => BoolV false)

	    | ev (Pair (e1,e2)) = PairV(ev e1,ev e2)
	    | ev (Fst(e)) = (case ev e 
			      of PairV(v1,v2) => v1)
	    | ev (Snd(e)) = (case ev e 
			      of PairV(v1,v2) => v2)
			    
	    | ev (EmptyC _) = SetV []
	    | ev (Singleton e) = SetV [ev e]
	    | ev (Union (e1,e2)) = 
	      (case (ev e1, ev e2) 
		of (SetV v1, SetV v2) => SetV (v1 @ v2))
	    (* TODO: Difference *)
	    | ev (Comp(e1,x,e2)) =  
	      (case ev e2
		of SetV(vs) => 
		   SetV (List.map (fn v => eval (Var.Map.insert(ctx,x,v)) e1)
				  vs))
	    | ev (Flatten e) = 
	      (case ev e 
		of SetV(vs) => SetV(List.concat (List.map (fn (SetV ws) => ws) vs)))
	    | ev (StringC s) = StringV s
	    | ev (Record (field_exps)) =  RecordV (Record.map ev field_exps)
	    | ev (Field (exp,lab)) =  
	      (case ev exp
		of RecordV (field_vals) =>
		   Record.lookup lab field_vals)
      in ev exp
      end


  fun eval_upd ctx stmt v = 
      let fun ev (Skip) v = v
	    | ev (Seq (s1,s2)) v = ev s2 (ev s1 v)
	    | ev (Repl exp) v = eval ctx exp
	    | ev (LetStmt(x,e,s)) v = 
	      let val v = eval ctx e 
	      in eval_upd (Var.Map.insert(ctx,x,v)) s v
	      end	    
	    | ev (IfThenElseStmt(e,s1,SOME s2)) v = 
	      (case (eval ctx e) 
		of (BoolV b) => if b then ev s1 v else ev s2 v)
	    | ev (IfThenElseStmt(e,s1,NONE)) v = 
	      (case (eval ctx  e) 
		of (BoolV b) => if b then ev s1 v else v)
	    | ev (Snapshot(x,s)) v = 
	      eval_upd (Var.Map.insert(ctx,x,v)) stmt v
	    | ev (UpdFst stmt) (PairV (v1,v2)) = PairV(ev stmt v1,v2)
	    | ev (UpdSnd stmt) (PairV (v1,v2)) = PairV(v1,ev stmt v2)
	    | ev (Iter stmt) (SetV vs) = SetV (List.map (ev stmt) vs)
	    | ev (Insert exp) (SetV vs) = 
	      (case eval ctx exp
		of SetV vs' => SetV (vs @ vs'))
	    | ev (Delete exp) (SetV vs) = 
	      (case eval ctx exp
		of SetV vs' => SetV (List.filter (fn v => not(List.exists (fn v' => v = v') vs')) vs))
	    | ev (UpdField (f,stmt)) (RecordV fvs) = 
	      RecordV (Record.fmap (fn (f',v') => if f = f' then ev stmt v' else v') fvs)
	    | ev (Add (f,exp)) (RecordV fvs) = 
	      RecordV(fvs@[(f,eval ctx exp)])
	    | ev (Drop f) (RecordV fvs) = 
	      RecordV(List.filter (fn (f',v) => f <> f') fvs)
      in ev stmt v
      end

end
