structure NRC = 
  struct


  open Record

  type var = Var.var

  datatype ty = BoolTy
	      | StringTy
	      | IntTy 
	      | PairTy of ty * ty 
	      | SetTy of ty
	      | RecordTy of ty record

  fun equiv_ty (BoolTy, BoolTy) = true
    | equiv_ty (StringTy, StringTy) = true
    | equiv_ty (IntTy, IntTy) = true
    | equiv_ty (PairTy(ty1,ty2), PairTy(ty1',ty2')) = 
      equiv_ty (ty1, ty1') andalso equiv_ty (ty2, ty2')
    | equiv_ty (SetTy(ty), SetTy(ty')) =  equiv_ty (ty, ty')
    | equiv_ty (RecordTy ty, RecordTy ty') = Record.equiv equiv_ty (ty, ty')
    | equiv_ty (_, _) = false


  datatype exp = Var of var
	       | Let of var * exp * exp
	       | Pair of exp * exp
	       | Fst of exp 
	       | Snd of exp
	       | IntC of int
	       | Plus of exp * exp
	       | Times of exp * exp
	       | Sum of exp
	       | Max of exp
	       | Min of exp
	       | Empty of exp
	       | Equal of exp * exp
	       | BoolC of bool
	       | Neg of exp
	       | And of exp * exp
	       | IfThenElse of exp * exp * exp option
	       | EmptyC of ty
	       | Singleton of exp
	       | Union of exp * exp
	       | Diff of exp * exp
	       | Comp of exp * var * exp
	       | Flatten of exp
	       | StringC of string
	       | Record of exp record
	       | Field of exp * string

    type ctx = ty Var.Map.map;


    fun swap0 (a,b) c = if Var.eq (a,c) then b
	                else if Var.eq (b,c) then a
                        else c

    fun swap (a,b) exp = 
      let fun h (Var(v)) = Var(swap0 (a,b) v)
        | h (Let(v,e1,e2)) = Let(swap0 (a,b) v, h e1, h e2)
	| h (Pair(e1,e2)) = Pair(h e1, h e2)
        | h (Fst(e)) = Fst(h e)
        | h (Snd(e)) = Snd(h e)
        | h (IntC(i)) = IntC(i)
        | h (Plus(e1,e2)) = Plus(h e1, h e2)
        | h (Times(e1,e2)) = Times(h e1, h e2)
        | h (Sum(e)) = Sum(h e)
        | h (Max(e)) = Max(h e)
        | h (Min(e)) = Min(h e)
        | h (Empty exp) = Empty (h exp)
        | h (Equal(e1,e2)) = Equal(h e1, h e2)
        | h (BoolC(b)) = BoolC(b)
        | h (Neg(e)) = Neg(h e)
        | h (And(e1,e2)) = And(h e1, h e2)
        | h (IfThenElse(e1,e2,SOME e3)) = IfThenElse(h e1,h e2,SOME(h e3))
        | h (IfThenElse(e1,e2,NONE)) = IfThenElse(h e1,h e2,NONE)
        | h (EmptyC ty) = EmptyC ty
        | h (Singleton(e)) = Singleton(h e)
        | h (Union(e1,e2)) = Union(h e1, h e2)
        | h (Diff(e1,e2)) = Diff(h e1, h e2)
        | h (Comp(e1,x,e2)) = Comp(h e1, swap0(a,b) x, h e2)
        | h (Flatten(e)) = Flatten(h e)
	| h (StringC(s)) = StringC(s)
	| h (Record (field_exps)) = 
	  Record (List.map (fn (f,e) => (f,h e)) field_exps)
	| h (Field(e,a)) = Field(h e, a)
      in 
        h exp 
      end


    fun infer_ty (ctx,exp) = 
      let fun infer (Var(v)) = 
            (case Var.Map.find(ctx,v)
               of SOME ty => ty)
        | infer (Let(v,e1,e2)) = 
            let val ty = infer e1
            in
              infer_ty (Var.Map.insert(ctx,v,ty), e2)
            end
	| infer (Pair(e1,e2)) = PairTy(infer e1, infer e2)
        | infer (Fst(e)) = (case infer e of PairTy(t1,_) => t1)
        | infer (Snd(e)) = (case infer e of PairTy(_,t2) => t2)
        | infer (IntC(i)) = IntTy
        | infer (Plus(e1,e2)) = 
            (case (infer e1, infer e2) of (IntTy,IntTy) => IntTy)
        | infer (Times(e1,e2)) = 
            (case (infer e1, infer e2) of (IntTy,IntTy) => IntTy)
        | infer (Sum(e)) = 
            (case (infer e) of SetTy(IntTy) => IntTy)
        | infer (Max(e)) = 
            (case (infer e) of SetTy(IntTy) => IntTy)
        | infer (Min(e)) = 
            (case (infer e) of SetTy(IntTy) => IntTy)
        | infer (Empty(e)) = 
            (case (infer e) of SetTy(_) => BoolTy)
        | infer (Equal(e1,e2)) = if infer(e1) = infer(e2) then BoolTy else raise Match
        | infer (BoolC(b)) = BoolTy
        | infer (Neg(e)) = (case infer e of BoolTy => BoolTy)
        | infer (And(e1,e2)) = (case (infer e1, infer e2) of (BoolTy, BoolTy) => BoolTy)
        | infer (IfThenElse(e1,e2,SOME(e3))) = 
	    (case infer e1 of BoolTy => 
            let val ty1 = infer e2
                val ty2 = infer e3 
            in if ty1 = ty2 then ty1 else raise Match
            end)
        | infer (IfThenElse(e1,e2,NONE)) = 
	    (case infer e1 of BoolTy => 
            let val ty1 = infer e2
                
            in case ty1 
		of SetTy(_) => ty1
		 | _ => raise Match
	    end)
        | infer (EmptyC ty) = SetTy ty
        | infer (Singleton(e)) = SetTy(infer e)
        | infer (Union(e1,e2)) = 
            (case (infer e1, infer e2) 
              of (SetTy(ty1),SetTy(ty2)) => if ty1 = ty2 then SetTy(ty1) else raise Match)
        | infer (Diff(e1,e2)) =             
            (case (infer e1, infer e2) 
              of (SetTy(ty1),SetTy(ty2)) => if ty1 = ty2 then SetTy(ty1) else raise Match)
        | infer (Comp(e1,v,e2)) = 
            (case infer e2 
              of SetTy(ty) => SetTy(infer_ty (Var.Map.insert(ctx,v,ty),e1)))
        | infer (Flatten(e)) = 
            (case infer e
              of SetTy(SetTy(ty)) => SetTy(ty))
	| infer (StringC(s)) = StringTy
	| infer (Record(field_exps)) = 
	  RecordTy(List.map (fn (f,e) => (f,infer e)) field_exps)
	| infer (Field(e,f)) = 
	  (case infer e 
		of RecordTy(field_tys) => 
		   Record.lookup f field_tys)
		 
    in infer exp 
    end
    and check_ty (ctx,exp,ty) = equiv_ty (ty, infer_ty (ctx,exp))

  (* Put into A-normal form, freshen all bound var's *)
    fun normalize (Var v) = Var v
      | normalize (Let (x,e1,e2)) = 
	let val x' = Var.fresh_var "x" 
	in Let(x',normalize e1, normalize (swap (x,x') e2))
	end
      | normalize (Pair(e1,e2)) = 
	let val x1 = Var.fresh_var("x") 
	    val x2 = Var.fresh_var("x")
	in
	    Let(x1,normalize e1,Let(x2,normalize e2,Pair(Var x1, Var x2)))
	end
      | normalize (Fst e) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Fst(Var x))
	end
      | normalize(Snd e) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Snd(Var x))
	end
      | normalize (IntC i) = IntC i
      | normalize (Plus (e1,e2)) = 
	let val x1 = Var.fresh_var("x") 
	    val x2 = Var.fresh_var("x")
	in
	    Let(x1,normalize e1,Let(x2,normalize e2,Plus(Var x1, Var x2)))
	end
      | normalize (Times (e1,e2)) = 
	let val x1 = Var.fresh_var("x") 
	    val x2 = Var.fresh_var("x")
	in
	    Let(x1,normalize e1,Let(x2,normalize e2,Times(Var x1, Var x2)))
	end
      | normalize (Sum(e)) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Sum(Var x))
	end
      | normalize (Min(e)) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Min(Var x))
	end
      | normalize (Max(e)) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Max(Var x))
	end
      | normalize (Empty (e)) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Empty(Var x))
	end
      | normalize (Equal (e1,e2)) = 
	let val x1 = Var.fresh_var("x") 
	    val x2 = Var.fresh_var("x")
	in
	    Let(x1,normalize e1,Let(x2,normalize e2,Equal(Var x1, Var x2)))
	end
      | normalize (BoolC b) = BoolC b
      | normalize (Neg e) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Neg(Var x))
	end
      | normalize (And (e1,e2)) = 
	let val x1 = Var.fresh_var("x") 
	    val x2 = Var.fresh_var("x")
	in
	    Let(x1,normalize e1,Let(x2,normalize e2,And(Var x1, Var x2)))
	end
      | normalize (IfThenElse (e0,e1,e2)) = 
	let val x = Var.fresh_var("x") 
	    fun normalize_opt NONE = NONE
	      | normalize_opt (SOME e) = SOME (normalize e)
	in
	    Let(x,normalize e0,IfThenElse(Var x,normalize e1, normalize_opt e2))
	end
      | normalize (EmptyC ty) = EmptyC ty
      | normalize (Singleton (e)) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Singleton(Var x))
	end
      | normalize (Union (e1,e2)) = 
	let val x1 = Var.fresh_var("x") 
	    val x2 = Var.fresh_var("x")
	in
	    Let(x1,normalize e1,Let(x2,normalize e2,Union(Var x1, Var x2)))
	end
      | normalize (Diff (e1,e2)) = 
	let val x1 = Var.fresh_var("x") 
	    val x2 = Var.fresh_var("x")
	in
	    Let(x1,normalize e1,Let(x2,normalize e2,Diff(Var x1, Var x2)))
	end
      | normalize (Comp (e,x,e0)) = 
	let val x' = Var.fresh_var "x" 
	    val x0 = Var.fresh_var("x") 
	in
	    Let(x0,normalize e0,Comp(normalize (swap (x,x') e),x',Var x0))
	end
      | normalize (Flatten (e)) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Flatten(Var x))
	end   
      | normalize (StringC s) = StringC s
      | normalize (Record r) = 
	let fun norm_rec [] = (fn e => e, [])
	      | norm_rec ((a,e)::r) = 
		let val x = Var.fresh_var("x") 
		    val (f,r') = norm_rec r
		in 
		    (fn e' => Let (x,normalize e,f e'), (a,Var x)::r')
		end
	    val (f,r') = norm_rec r 
	in f (Record r')
	end
      | normalize (Field (e,a)) = 
	let val x = Var.fresh_var("x") 
	in
	    Let(x,normalize e,Field(Var x,a))
	end  

    fun toString_ty BoolTy = "bool"
      | toString_ty StringTy = "string"
      | toString_ty IntTy = "int"
      | toString_ty (PairTy (ty1,ty2)) = toString_ty ty1^" * "^ toString_ty ty2
      | toString_ty (SetTy ty) = "{"^toString_ty ty^"}"
      | toString_ty (RecordTy r) = 
	"<"^String.concatWith "," (List.map (fn (a,ty) => a^":"^toString_ty ty) r)^">"
		


    fun toString e = 
	case e  
	 of Var v => Var.toString v
	  | Let(v,e1,e2) 
	    => "let "^Var.toString v^ " = "^toString e1^ " in "^ toString e2
	  | Pair(e1,e2)
	    => "<"^toString e1^","^ toString e2^">"
	  | Fst(e)
	    => "fst("^toString e^")"
	  | Snd(e)
	    => "snd("^toString e^")"
	  | IntC(i) => Int.toString i
	  | Plus(e1,e2)
	    => toString e1^" + "^ toString e2
	  | Times(e1,e2)
	    => toString e1^" * "^ toString e2
	  | Sum (e)
	    => "sum("^toString e^")"
	  | Max (e)
	    => "max("^toString e^")"
	  | Min (e)
	    => "min("^toString e^")"
	  | Empty (e)
	    => "empty("^toString e^")"
	  | Equal (e1,e2)
	    => toString e1^" = "^ toString e2
	  | BoolC (b) => Bool.toString b
	  | Neg (e)
	    => "sum("^toString e^")"
	  | And (e1,e2)
	    => toString e1^" /\\ "^ toString e2
	  | IfThenElse (e0,e1,SOME e2) 
	    => "if "^toString e0^" then "^toString e1^" else"^toString e2
	  | IfThenElse (e0,e1,NONE) 
	    => "if "^toString e0^" then "^toString e1
	  | EmptyC (ty) => "{|"^toString_ty ty^"|}"
	  | Singleton (e)=> "{"^toString e^"}"
	  | Union (e1,e2)
	    => toString e1^" u "^ toString e2
	  | Diff (e1,e2)
	    => toString e1^" \\ "^ toString e2
	  | Comp (e,x,e0)
	    => "{"^toString e^ " | "^Var.toString x^ " <- "^ toString e0^"}"
	  | Flatten (e) => "U "^toString e
	  | StringC (s) => "\""^s^"\""
	  | Record (r) 
	    => "<"^String.concatWith "," (List.map (fn (a,e) => a^":"^toString e) r)^">"
	
	  | Field (e,a) => toString e^"."^a


    fun latex_ty BoolTy = "bool"
      | latex_ty StringTy = "string"
      | latex_ty IntTy = "int"
      | latex_ty (PairTy (ty1,ty2)) = latex_ty ty1^" \\times "^ latex_ty ty2
      | latex_ty (SetTy ty) = "\\{"^latex_ty ty^"\\}"
      | latex_ty (RecordTy r) = 
	"\\langle"^String.concatWith "," (List.map (fn (a,ty) => a^":"^latex_ty ty) r)^"\\rangle"

    fun latex_map latex_a ctx = 
        "["^String.concatWith "," (Var.Map.foldli (fn(k,v,ss) => Var.latex_var k ^":="^latex_a v::ss) [] ctx)^"]"

    fun latex_exp e = 
	case e  
	 of Var v => Var.latex_var v
	  | Let(v,e1,e2) 
	    => "let\\ "^Var.latex_var v^ " = "^latex_exp e1^ "\\ in\\ "^ latex_exp e2
	  | Pair(e1,e2)
	    => "("^latex_exp e1^","^ latex_exp e2^")"
	  | Fst(e)
	    => "fst("^latex_exp e^")"
	  | Snd(e)
	    => "snd("^latex_exp e^")"
	  | IntC(i) => Int.toString i
	  | Plus(e1,e2)
	    => latex_exp e1^" + "^ latex_exp e2
	  | Times(e1,e2)
	    => latex_exp e1^" * "^ latex_exp e2
	  | Sum (e)
	    => "sum("^latex_exp e^")"
	  | Max (e)
	    => "max("^latex_exp e^")"
	  | Min (e)
	    => "min("^latex_exp e^")"
	  | Empty (e)
	    => "empty("^latex_exp e^")"
	  | Equal (e1,e2)
	    => latex_exp e1^" = "^ latex_exp e2
	  | BoolC (b) => Bool.toString b
	  | Neg (e)
	    => "sum("^latex_exp e^")"
	  | And (e1,e2)
	    => latex_exp e1^" \\wedge "^ latex_exp e2
	  | IfThenElse (e0,e1,SOME e2) 
	    => "if\\ "^latex_exp e0^"\\ then\\ "^latex_exp e1^"\\ else\\ "^latex_exp e2
	  | IfThenElse (e0,e1,NONE) 
	    => "if\\ "^latex_exp e0^"\\ then\\ "^latex_exp e1
	  | EmptyC (ty) => "\\emptyset_{"^latex_ty ty^"}"
	  | Singleton (e)=> "\\{"^latex_exp e^"\\}"
	  | Union (e1,e2)
	    => latex_exp e1^" \\cup "^ latex_exp e2
	  | Diff (e1,e2)
	    => latex_exp e1^" \\backslash "^ latex_exp e2
	  | Comp (e,x,e0)
	    => "\\{"^latex_exp e^ " \\mid "^Var.latex_var x^ " \\leftarrow "^ latex_exp e0^"\\}"
	  | Flatten (e) => "\\bigcup "^latex_exp e
	  | StringC (s) => "\""^s^"\""
	  | Record (r) 
	    => "\\langle"^String.concatWith "," (List.map (fn (a,e) => a^":"^latex_exp e) r)^"\\rangle"
	
	  | Field (e,a) => latex_exp e^"."^a

        

  datatype stmt = 
	   Repl of exp
	 | Snapshot of var * stmt
	 | LetStmt of var * exp * stmt 
	 | Skip
	 | Seq of stmt * stmt
	 | IfThenElseStmt of exp * stmt * stmt option

	 | UpdFst of stmt
	 | UpdSnd of stmt

	 | Iter of stmt
	 | Insert of exp
	 | Delete of exp

	 | UpdField of string * stmt
	 | Add of string * exp
	 | Drop of string
	
(* should go somewhere else! *)
  datatype start = Exp of exp | Stmt of stmt

	  
    fun infer_stmt_ty (ctx,stmt,ty) = 
      let fun infer (Skip,ty) = ty
	    | infer (Seq(s1,s2),ty) = 
	      infer(s2,infer(s1,ty))
	    | infer (Repl(exp),ty) = infer_ty (ctx,exp)
	    | infer (Snapshot(v,stmt),ty) = 
	      infer_stmt_ty (Var.Map.insert(ctx,v,ty), stmt,ty)
	    | infer (LetStmt(v,exp,stmt),ty) = 
	      infer_stmt_ty (Var.Map.insert(ctx,v,infer_ty(ctx,exp)), stmt,ty)
	    | infer (IfThenElseStmt(exp,stmt1,SOME stmt2),ty) = 
		     (case infer_ty(ctx,exp) 
		       of BoolTy => 
			  let val ty1 = infer(stmt1,ty)
			      val ty2 = infer(stmt2,ty)
			  in
			      if equiv_ty (infer(stmt1,ty),infer(stmt2,ty))
			      then ty1
			      else raise Match
			  end
		     ) 
	    | infer (IfThenElseStmt(exp,stmt1,NONE),ty) = 
		     (case infer_ty(ctx,exp) 
		       of BoolTy => infer(stmt1,ty))
			    
	    | infer(UpdFst stmt,PairTy(ty1,ty2)) = PairTy(infer(stmt,ty1),ty2)
	    | infer(UpdSnd stmt,PairTy(ty1,ty2)) = PairTy(ty1,infer(stmt,ty2))

	    | infer(UpdField (lab,stmt),RecordTy(rty)) = 
	      RecordTy(Record.fmap (fn (f,ty) => 
				       if f = lab 
				       then infer(stmt,ty)
				       else ty)
				  rty)
	    | infer(Add(lab,exp),RecordTy(rty)) = 
	      if (List.exists (fn (f,_) => f = lab) rty)
	      then raise Match
	      else RecordTy(rty@[(lab,infer_ty(ctx,exp))])
	    | infer(Drop(lab),RecordTy(rty)) = 
	      if (List.exists (fn (f,_) => f = lab) rty)
	      then RecordTy(List.filter (fn (f,ty) => f <> lab) rty)
	      else raise Match
	      
	    | infer(Iter stmt,SetTy(ty)) = SetTy(infer(stmt,ty))
	    | infer(Insert exp, SetTy(ty)) = 
	      if check_ty(ctx,exp,SetTy(ty))
	      then SetTy(ty)
	      else raise Match
	    | infer(Delete exp, SetTy(ty)) = 
	      if check_ty(ctx,exp,SetTy(ty))
	      then SetTy(ty)
	      else raise Match

      in infer(stmt,ty)
      end




  end
