functor TaggedEvalWhereFun(structure TaggedValue : TAGGED_VALUE) = struct
  (* Standard evaluation *)			 
  open NRC;
  open TaggedValue;
  open TaggedValue.Tag;
  exception EvalFailure


  fun eval ctx exp = 
      let fun ev (Var(x)) = (case Var.Map.find(ctx,x) 
			      of SOME v => v)
	    | ev (Let(x,e1,e2)) = 
	      let val v = ev e1 
	      in eval (Var.Map.insert(ctx,x,v)) e2
	      end

	    | ev (BoolC c) = Tag(empty,BoolV c)
	    | ev (And(e1,e2)) = 
	      (case (ev e1, ev e2) 
		of (Tag(t1,BoolV b1), Tag(t2,BoolV b2)) => 
		   Tag(empty,BoolV (b1 andalso b2))
		 | _ => raise Match)
	    | ev (Neg(e)) = 
	      (case (ev e) 
		of (Tag(t,BoolV b)) => Tag(empty,BoolV (not(b)))
		 | _ => raise Match)
	    | ev (IfThenElse(e,e1,SOME e2)) = 
	      (case (ev e) 
		of (Tag(t,BoolV b)) => 
		   if b
		   then ev e1 
		   else ev e2 
		 | _ => raise Match)
	    | ev (IfThenElse(e,e1,NONE)) = 
	      (case (ev e) 
		of (Tag(t,BoolV b)) => 
		   if b
		   then ev e1 
		   else Tag(empty,SetV([])) 
		 | _ => raise Match)
	    | ev(Equal(e1,e2)) = 
	      let val v1 = ev e1 
		  val v2 = ev e2 
	      in
		  Tag(empty,BoolV(eq(v1,v2)))
	      end
	       	       
	    | ev (IntC c) = Tag(empty,IntV c)
	    | ev (Plus(e1,e2)) = 
	       (case (ev e1, ev e2) 
		of (Tag(t1,IntV i1), Tag(t2,IntV i2)) => 
		   Tag(empty,IntV (i1 + i2))
		 | _ => raise Match)
	    | ev (Times(e1,e2)) = 
	       (case (ev e1, ev e2) 
		of (Tag(t1,IntV i1), Tag(t2,IntV i2)) => 
		   Tag(empty,IntV (i1 * i2))
		 | _ => raise Match)
	    | ev (Sum(e)) = 
	       (case (ev e) 
		 of (Tag(t0,SetV vs)) => 
		    let val (s) = List.foldr 
					(fn (Tag(t,IntV i), (s)) => 
					   (i+s)) (0) vs
		   in Tag(empty,IntV s) 
		   end
		 | _ => raise Match)
	    | ev (Min(e)) = 
	      let val SOME maxInt = Int.maxInt 
	      in 
	       (case (ev e) 
		 of (Tag(t0,SetV vs)) => 
		    let val (s) = List.foldr 
					(fn (Tag(t,IntV i), (s)) => 
					   (Int.min(i,s))) (maxInt) vs
		   in Tag(empty,IntV s) 
		   end
		 | _ => raise Match)
	      end
	    | ev (Max(e)) = 
	      let val SOME minInt = Int.minInt 
	      in 
		  (case (ev e) 
		    of (Tag(t0,SetV vs)) => 
		       let val (s) = List.foldr 
					   (fn (Tag(t,IntV i), (s)) => 
					       (Int.max(i,s))) (minInt) vs
		       in Tag(empty,IntV s) 
		       end
		 | _ => raise Match)
	      end
	    | ev (Empty(e)) = 
	      (case (ev e) 
		of (Tag(t0,SetV [])) => Tag(empty,BoolV true)
		 | (Tag(t0,SetV _)) => Tag(empty,BoolV false)
		 | _ => raise Match)

	    | ev (Pair (e1,e2)) = Tag(empty,PairV(ev e1,ev e2))
	    | ev (Fst(e)) = (case ev e 
			      of (Tag(t,PairV(v1,_))) => v1
		 | _ => raise Match)
	    | ev (Snd(e)) = (case ev e 
			      of (Tag(t,PairV(_,v2))) => v2
		 | _ => raise Match)
	    | ev (EmptyC _) = Tag(empty,SetV [])
	    | ev (Singleton e) = Tag(empty,SetV [ev e])
	    | ev (Union (e1,e2)) = 
	      (case (ev e1, ev e2) 
		of (Tag(t1,SetV v1), Tag(t2,SetV v2)) => 
		   Tag(empty,SetV (v1 @ v2))
		 | _ => raise Match)
	    | ev (Diff(e1,e2)) = raise Match
	    | ev (Comp(e1,x,e2)) =  
	      (case ev e2
		of Tag(t,SetV(vs)) => 
		   Tag(t,SetV (List.map 
				   (fn v => eval (Var.Map.insert(ctx,x,v)) e1)
				  vs))
		 | _ => raise Match)

	    | ev (Flatten e) =  
	      (case ev e 
		of Tag(t0,SetV(vs)) =>
		   let fun concat (Tag(t,SetV(vs)),(vs')) = 
			   (vs@vs')
		   in 
		   Tag(empty,SetV(List.foldr concat ([]) vs))
		   end
		 | _ => raise Match)
	    | ev (StringC s) = Tag(empty,StringV s)
	    | ev (Record (field_exps)) = 
	      Tag(empty,RecordV(Record.map ev field_exps))
	    | ev (Field(e,lab)) = 
	      (case ev e 
		of (Tag(t,RecordV(r))) => 
		   Record.lookup lab r
		 | _ => raise Match)
      in ev exp
      end


  fun eval_upd ctx stmt v = 
      let fun ev (Skip) v = v
	    | ev (Seq (s1,s2)) v = ev s2 (ev s1 v)
	    | ev (Repl exp) v = eval ctx exp
	    | ev (Snapshot(x,s)) v = 
	      eval_upd (Var.Map.insert(ctx,x,v)) s v
	    | ev (LetStmt(x,e,s)) v = 
	      let val v = eval ctx e
	      in eval_upd (Var.Map.insert(ctx,x,v)) s v
	      end
	    | ev (IfThenElseStmt(e,s1,SOME s2)) v = 
	      (case eval ctx e 
		 of Tag(_,BoolV b) => if b then ev s1 v else ev s2 v)
	    | ev (IfThenElseStmt(e,s1,NONE)) v = 
	      (case eval ctx e 
		 of Tag(_,BoolV b) => if b then ev s1 v else v)

	    | ev (UpdFst stmt) (Tag(t,PairV (v1,v2))) = 
	      Tag(t,PairV(ev stmt v1,v2))
	    | ev (UpdSnd stmt) (Tag(t,PairV (v1,v2))) = 
	      Tag(t,PairV(v1,ev stmt v2))
	    | ev (Iter stmt) (Tag(c,SetV vs)) = 
	      Tag(c,SetV (List.map (ev stmt) vs))
	    | ev (Insert exp) (Tag(c,SetV vs)) = 
	      (case eval ctx exp
		of (Tag(_,SetV vs')) => Tag(c,SetV (vs @ vs')))
	    | ev (Delete exp) (Tag(c,SetV vs)) = 
	      (case eval ctx exp
		of (Tag(_,SetV vs')) => 
		   Tag(c,SetV (List.filter 
			     (fn v => not(List.exists 
					      (fn v' => erase v = erase v') 
					      vs')) 
			     vs)))
	    | ev (UpdField (f,stmt)) (Tag(c,RecordV fvs)) = 
	      Tag(c,RecordV 
			(Record.fmap 
			     (fn (f',v') => if f = f' 
					    then ev stmt v' 
					    else v') 
			     fvs))
	    | ev (Add (f,exp)) (Tag(c,RecordV fvs)) = 
	      Tag(c,RecordV(fvs@[(f,eval ctx exp)]))
	    | ev (Drop f) (Tag(c,RecordV fvs)) = 
	      Tag(c,RecordV(List.filter (fn (f',v) => f <> f') fvs))
      in ev stmt v
      end

end




(*
structure  TaggedEval = struct
  (* Standard evaluation *)			 
  open NRC
  open TaggedValue
  open Var.Set
  exception EvalFailure

  fun eval ctx exp = 
      let fun ev (Var(x)) = (case Var.Map.find(ctx,x) 
			      of SOME v => v)
	    | ev (Let(x,e1,e2)) = 
	      let val v = ev e1 
	      in eval (Var.Map.insert(ctx,x,v)) e2
	      end

	    | ev (BoolC c) = Tag(empty,BoolV c)
	    | ev (And(e1,e2)) = 
	      (case (ev e1, ev e2) 
		of (Tag(t1,BoolV b1), Tag(t2,BoolV b2)) => 
		   Tag(union(t1,t2),BoolV (b1 andalso b2))
		 | _ => raise Match)
	    | ev (Neg(e)) = 
	      (case (ev e) 
		of (Tag(t,BoolV b)) => Tag(t,BoolV (not(b)))
		 | _ => raise Match)
	    | ev (IfThenElse(e,e1,SOME e2)) = 
	      (case (ev e) 
		of (Tag(t,BoolV b)) => 
		   if b
		   then let val Tag(t1,v1) = ev e1 
			in Tag(union(t,t1),v1) 
			      end 
		   else let val Tag(t2,v2) = ev e2 
			in Tag(union(t,t2),v2) 
			end
		 | _ => raise Match)
	    | ev (IfThenElse(e,e1,NONE)) = 
	      (case (ev e) 
		of (Tag(t,BoolV b)) => 
		   if b
		   then let val Tag(t1,v1) = ev e1 
			in Tag(union(t,t1),v1) 
			      end 
		   else Tag(t,SetV([])) 
		 | _ => raise Match)
	    | ev(Equal(e1,e2)) = 
	      let val v1 = ev e1 
		  val v2 = ev e2 
	      in
		  Tag(union(prov(v1),prov(v2)),BoolV(eq(v1,v2)))
	      end
	       	       
	    | ev (IntC c) = Tag(empty,IntV c)
	    | ev (Plus(e1,e2)) = 
	       (case (ev e1, ev e2) 
		of (Tag(t1,IntV i1), Tag(t2,IntV i2)) => 
		   Tag(union(t1,t2),IntV (i1 + i2))
		 | _ => raise Match)
	    | ev (Times(e1,e2)) = 
	       (case (ev e1, ev e2) 
		of (Tag(t1,IntV i1), Tag(t2,IntV i2)) => 
		   Tag(union(t1,t2),IntV (i1 * i2))
		 | _ => raise Match)
	    | ev (Sum(e)) = 
	       (case (ev e) 
		 of (Tag(t0,SetV vs)) => 
		    let val (t,s) = List.foldr 
					(fn (Tag(t,IntV i), (t',s)) => 
					   (union(t,t'),i+s)) (t0,0) vs
		   in Tag(t,IntV s) 
		   end
		 | _ => raise Match)
	    | ev (Pair (e1,e2)) = Tag(empty,PairV(ev e1,ev e2))
	    | ev (Fst(e)) = (case ev e 
			      of (Tag(t,PairV(Tag(u,w1),_))) => 
				 Tag(union(t,u),w1)
		 | _ => raise Match)
	    | ev (Snd(e)) = (case ev e 
			      of (Tag(t,PairV(_,Tag(u,w2)))) => 
				 Tag(union(t,u),w2)
		 | _ => raise Match)
	    | ev (EmptyC _) = Tag(empty,SetV [])
	    | ev (Singleton e) = Tag(empty,SetV [ev e])
	    | ev (Union (e1,e2)) = 
	      (case (ev e1, ev e2) 
		of (Tag(t1,SetV v1), Tag(t2,SetV v2)) => 
		   Tag(union(t1,t2),SetV (v1 @ v2))
		 | _ => raise Match)
	    | ev (Diff(e1,e2)) = raise Match
	    | ev (Comp(e1,x,e2)) =  
	      (case ev e2
		of Tag(t,SetV(vs)) => 
		   Tag(t,SetV (List.map 
				   (fn v => eval (Var.Map.insert(ctx,x,v)) e1)
				  vs))
		 | _ => raise Match)

	    | ev (Flatten e) =  
	      (case ev e 
		of Tag(t0,SetV(vs)) =>
		   let fun concat (Tag(t,SetV(vs)),(t',vs')) = 
			   (union(t,t'),vs@vs')
		       val (t,set) = List.foldl concat (t0,[]) vs
		   in 
		   Tag(t,SetV(set))
		   end
		 | _ => raise Match)
	    | ev (StringC s) = notag(StringV s)
      in ev exp
      end

end
*)
